# sage_setup: distribution = sagemath-graphs
# distutils: language = c++
# distutils: extra_compile_args = -std=c++11
r"""
Modular Decomposition

This module implements the function for computing the modular decomposition
of undirected graphs.

AUTHORS:

- Lokesh Jain (2017): first implementation of the linear time algorithm of
  D. Corneil, M. Habib, C. Paul and M. Tedder [TCHP2008]_

- David Einstein (2018): added the algorithm of M. Habib and M. Maurer [HM1979]_

- Cyril Bouvier (2024): second implementation of the linear time algorithm
  of D. Corneil, M. Habib, C. Paul and M. Tedder [TCHP2008]_
"""
# ****************************************************************************
#       Copyright (C) 2017 Lokesh Jain <lokeshj1703@gmail.com>
#                     2018 David Einstein <deinst@gmail.com>
#                     2024 Cyril Bouvier <cyril.bouvier@lirmm.fr>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from cython.operator cimport dereference as deref

from enum import IntEnum

from sage.graphs.base.c_graph cimport CGraph, CGraphBackend
from sage.graphs.graph_decompositions.slice_decomposition cimport \
        extended_lex_BFS
from sage.misc.random_testing import random_testing

try:
    from sage.groups.perm_gps.permgroup_element import PermutationGroupElement
except ImportError:
    PermutationGroupElement = ()


################################################################################
#                     Corneil-Habib-Paul-Tedder algorithm                      #
################################################################################
def corneil_habib_paul_tedder_algorithm(G):
    r"""
    Compute the modular decomposition by the algorithm of Corneil, Habib, Paul
    and Tedder.

    INPUT:

    - ``G`` -- the graph for which modular decomposition tree needs to be
      computed

    OUTPUT: an object of type Node representing the modular decomposition tree
    of the graph G

    This function computes the modular decomposition of the given graph by the
    algorithm of Corneil, Habib, Paul and Tedder [TCHP2008]_. It is a recursive,
    linear-time algorithm that first computes the slice decomposition of the
    graph (via the extended lexBFS algorithm) and then computes the modular
    decomposition by calling itself recursively on the slices of the previously
    computed slice decomposition.

    This functions is based on the last version of the paper [TCHP2008]_.
    Previous versions of the paper and previous implementations were found to
    contains errors, see [AP2024]_.

    .. SEEALSO::

        * :mod:`~sage.graphs.graph_decompositions.slice_decomposition` --
          compute a slice decomposition of the simple undirect graph

    This function should not be used directly, it should be called via the
    ``modular_decomposition`` method of ``Graph`` with the parameter
    ``algorithm='corneil_habib_paul_tedder'``.

    This functions assumes that ``graph`` is a object of the class ``Graph`` and
    is a simple graph.

    TESTS::

        sage: from sage.graphs.graph_decompositions.modular_decomposition import *
        sage: recreate_decomposition(15, corneil_habib_paul_tedder_algorithm,
        ....:                                                         3, 4, 0.2)
        sage: recreate_decomposition(10, corneil_habib_paul_tedder_algorithm,
        ....:                                                         4, 5, 0.2)
        sage: recreate_decomposition(3, corneil_habib_paul_tedder_algorithm,
        ....:                                                         6, 5, 0.2)

        sage: H = Graph('Hv|mmjz', format='graph6')
        sage: H.relabel('abcdefghi')  # counter-exemple graph from [AP2024]_
        sage: H.modular_decomposition()
        (SERIES,
         [(PRIME, ['e', (PARALLEL, ['g', 'h']), 'b', 'c']),
          (PARALLEL, [(SERIES, ['i', 'd', 'a']), 'f'])])
    """
    cdef CGraphBackend Gbackend = <CGraphBackend> G._backend
    cdef CGraph cg = Gbackend.cg()

    cdef vector[int] sigma
    cdef vector[vector[int]] lex_label
    cdef vector[size_t] xslice_len

    # Compute the slice decomposition using the extended lexBFS algorithm
    extended_lex_BFS(cg, sigma, NULL, -1, NULL, &xslice_len, &lex_label)

    cdef SDData SD
    SD.set_from_data(0, sigma.data(), xslice_len.data(), lex_label.data())

    MD = corneil_habib_paul_tedder_inner(SD)

    r = md_tree_node_to_md_tree(MD, Gbackend)
    dealloc_md_tree_nodes_recursively(MD)
    return r


cdef object _md_tree_node_to_md_tree_inner_rec(const md_tree_node *n,
                                               CGraphBackend Gb):
    """
    Utility function for :func:`md_tree_node_to_md_tree`.
    """
    cdef md_tree_node *c
    if deref(n).is_leaf():
        return Node.create_leaf(Gb.vertex_label(deref(n).vertex))

    if deref(n).is_series():
        node = Node(NodeType.SERIES)
    elif deref(n).is_parallel():
        node = Node(NodeType.PARALLEL)
    else:  # is_prime
        node = Node(NodeType.PRIME)
    node.children.extend(_md_tree_node_to_md_tree_inner_rec(c, Gb)
                                                    for c in deref(n).children)
    return node


cdef object md_tree_node_to_md_tree(const md_tree_node *n, CGraphBackend Gb):
    """
    This function converts a modular decomposition tree (given as a pointer to a
    md_tree_node) into an object of the Python class Node (which is then
    converted into the correct output by :meth:`Graph.modular_decomposition`).

    The graph backend is needed to convert the int stored into the md_tree_node
    into the corresponding vertex of the Python graph.

    This function deals with the case of an empty tree and then delegates the
    actual conversion to :func:`_md_tree_node_to_md_tree_inner_rec`.

    TESTS:

    Indirect doctests::

        sage: from sage.graphs.graph_decompositions.modular_decomposition import *
        sage: corneil_habib_paul_tedder_algorithm(Graph(1))
        NORMAL [0]
        sage: corneil_habib_paul_tedder_algorithm(Graph(2))
        PARALLEL [NORMAL [0], NORMAL [1]]
        sage: corneil_habib_paul_tedder_algorithm(graphs.CompleteGraph(3))
        SERIES [NORMAL [1], NORMAL [2], NORMAL [0]]
    """
    if n == NULL:
        return Node(NodeType.EMPTY)
    return _md_tree_node_to_md_tree_inner_rec(n, Gb)


################################################################################
class NodeType(IntEnum):
    """
    NodeType is an enumeration class used to define the various types of nodes
    in modular decomposition tree.

    The various node types defined are

    - ``PARALLEL`` -- indicates the node is a parallel module

    - ``SERIES`` -- indicates the node is a series module

    - ``PRIME`` -- indicates the node is a prime module

    - ``EMPTY`` -- indicates a empty tree

    - ``NORMAL`` -- indicates the node is normal containing a vertex
    """
    PRIME = 0
    SERIES = 1
    PARALLEL = 2
    NORMAL = 3
    EMPTY = -1

    def __repr__(self) -> str:
        r"""
        Return a string representation of a ``NodeType`` object.

        TESTS::

            sage: from sage.graphs.graph_decompositions.modular_decomposition import NodeType
            sage: repr(NodeType.PARALLEL)
            'PARALLEL'
            sage: str(NodeType.PRIME)
            'PRIME'
        """
        return self.name

    __str__ = __repr__


class Node:
    """
    Node class stores information about the node type.

    Node type can be ``PRIME``, ``SERIES``, ``PARALLEL``, ``NORMAL`` or
    ``EMPTY``.

    - ``node_type`` -- is of type NodeType and specifies the type of node
    """
    def __init__(self, node_type):
        r"""
        Create a node with the given node type.

        EXAMPLES::

            sage: from sage.graphs.graph_decompositions.modular_decomposition import *
            sage: n = Node(NodeType.SERIES); n.node_type
            SERIES
            sage: n.children
            []
        """
        self.node_type = node_type
        self.children = []

    def is_prime(self):
        r"""
        Check whether ``self`` is a prime node.

        EXAMPLES::

            sage: from sage.graphs.graph_decompositions.modular_decomposition import *
            sage: n = Node(NodeType.PRIME)
            sage: n.children.append(Node.create_leaf(1))
            sage: n.children.append(Node.create_leaf(2))
            sage: n.is_prime()
            True
            sage: (n.children[0].is_prime(), n.children[1].is_prime())
            (False, False)
        """
        return self.node_type == NodeType.PRIME

    def is_series(self):
        r"""
        Check whether ``self`` is a series node.

        EXAMPLES::

            sage: from sage.graphs.graph_decompositions.modular_decomposition import *
            sage: n = Node(NodeType.SERIES)
            sage: n.children.append(Node.create_leaf(1))
            sage: n.children.append(Node.create_leaf(2))
            sage: n.is_series()
            True
            sage: (n.children[0].is_series(), n.children[1].is_series())
            (False, False)
        """
        return self.node_type == NodeType.SERIES

    def is_empty(self):
        r"""
        Check whether ``self`` is an empty node.

        EXAMPLES::

            sage: from sage.graphs.graph_decompositions.modular_decomposition import *
            sage: Node(NodeType.EMPTY).is_empty()
            True
            sage: Node.create_leaf(1).is_empty()
            False
        """
        return self.node_type == NodeType.EMPTY

    def is_leaf(self):
        r"""
        Check whether ``self`` is a leaf.

        EXAMPLES::

            sage: from sage.graphs.graph_decompositions.modular_decomposition import *
            sage: n = Node(NodeType.PRIME)
            sage: n.children.append(Node.create_leaf(1))
            sage: n.children.append(Node.create_leaf(2))
            sage: n.is_leaf()
            False
            sage: all(c.is_leaf() for c in n.children)
            True
        """
        return self.node_type == NodeType.NORMAL

    def __repr__(self):
        r"""
        Return a string representation of the node.

        EXAMPLES::

            sage: from sage.graphs.graph_decompositions.modular_decomposition import *
            sage: n = Node(NodeType.PRIME)
            sage: n.children.append(Node.create_leaf(1))
            sage: n.children.append(Node.create_leaf(2))
            sage: str(n)
            'PRIME [NORMAL [1], NORMAL [2]]'
        """
        return f"{self.node_type} {self.children}"

    def __eq__(self, other):
        r"""
        Compare two nodes for equality.

        EXAMPLES::

            sage: from sage.graphs.graph_decompositions.modular_decomposition import *
            sage: n1 = Node(NodeType.PRIME)
            sage: n2 = Node(NodeType.PRIME)
            sage: n3 = Node(NodeType.SERIES)
            sage: n1 == n2
            True
            sage: n1 == n3
            False
        """
        return (self.node_type == other.node_type and
                self.children == other.children)

    @classmethod
    def create_leaf(cls, v):
        """
        Return Node object that is a leaf corresponding to the vertex ``v``.

        INPUT:

        - ``vertex`` -- vertex number

        OUTPUT: a node object representing the vertex with ``node_type`` set as
        ``NodeType.NORMAL``

        EXAMPLES::

            sage: from sage.graphs.graph_decompositions.modular_decomposition import Node
            sage: node = Node.create_leaf(2)
            sage: node
            NORMAL [2]
        """
        node = cls(NodeType.NORMAL)
        node.children.append(v)
        return node


def print_md_tree(root):
    """
    Print the modular decomposition tree.

    INPUT:

    - ``root`` -- root of the modular decomposition tree

    EXAMPLES::

        sage: from sage.graphs.graph_decompositions.modular_decomposition import *
        sage: print_md_tree(habib_maurer_algorithm(graphs.IcosahedralGraph()))
        PRIME
         3
         4
         7
         9
         11
         1
         5
         8
         0
         2
         6
         10
    """

    def recursive_print_md_tree(root, level):
        """
        Print the modular decomposition tree at root.

        INPUT:

        - ``root`` -- root of the modular decomposition tree

        - ``level`` -- indicates the depth of root in the original modular
          decomposition tree
        """
        if root.node_type != NodeType.NORMAL:
            print("{}{}".format(level, str(root.node_type)))
            for tree in root.children:
                recursive_print_md_tree(tree, level + " ")
        else:
            print("{}{}".format(level, str(root.children[0])))

    recursive_print_md_tree(root, "")


# =============================================================================
# Habib Maurer algorithm
# =============================================================================

def gamma_classes(graph):
    """
    Partition the edges of the graph into Gamma classes.

    Two distinct edges are Gamma related if they share a vertex but are not
    part of a triangle.  A Gamma class of edges is a collection of edges such
    that any edge in the class can be reached from any other by a chain of
    Gamma related edges (that are also in the class).

    The two important properties of the Gamma class

    * The vertex set corresponding to a Gamma class is a module
    * If the graph is not fragile (neither it or its complement is
      disconnected) then there is exactly one class that visits all the
      vertices of the graph, and this class consists of just the edges that
      connect the maximal strong modules of that graph.

    EXAMPLES:

    The gamma_classes of the octahedral graph are the three 4-cycles
    corresponding to the slices through the center of the octahedron::

        sage: from sage.graphs.graph_decompositions.modular_decomposition import gamma_classes
        sage: g = graphs.OctahedralGraph()
        sage: sorted(gamma_classes(g), key=str)
        [frozenset({0, 1, 4, 5}), frozenset({0, 2, 3, 5}), frozenset({1, 2, 3, 4})]

    TESTS:

    Ensure that the returned vertex sets from some random graphs are modules::

        sage: from sage.graphs.graph_decompositions.modular_decomposition import test_gamma_modules
        sage: test_gamma_modules(2, 10, 0.5)
    """
    from itertools import chain
    from sage.sets.disjoint_set import DisjointSet

    pieces = DisjointSet(frozenset(e) for e in graph.edge_iterator(labels=False))
    for v in graph:
        neighborhood = graph.subgraph(vertices=graph.neighbors(v))
        for component in neighborhood.complement().connected_components(sort=False):
            v1 = component[0]
            e = frozenset([v1, v])
            for vi in component[1:]:
                ei = frozenset([vi, v])
                pieces.union(e, ei)
    return {frozenset(chain.from_iterable(loe)): loe for loe in pieces}


def habib_maurer_algorithm(graph, g_classes=None):
    """
    Compute the modular decomposition by the algorithm of Habib and Maurer.

    Compute the modular decomposition of the given graph by the algorithm of
    Habib and Maurer [HM1979]_ . If the graph is disconnected or its complement
    is disconnected return a tree with a ``PARALLEL`` or ``SERIES`` node at the
    root and children being the modular decomposition of the subgraphs induced
    by the components. Otherwise, the root is ``PRIME`` and the modules are
    identified by having identical neighborhoods in the gamma class that spans
    the vertices of the subgraph (exactly one is guaranteed to exist). The gamma
    classes only need to be computed once, as the algorithm computes the the
    classes for the current root and each of the submodules. See also [BM1983]_
    for an equivalent algorithm described in greater detail.

    This function should not be used directly, it should be called via the
    ``modular_decomposition`` method of ``Graph`` with the parameter
    ``algorithm='habib_maurer'``.

    This functions assumes that ``graph`` is a object of the class ``Graph``, is
    a simple graph and has at least 1 vertex.

    INPUT:

    - ``graph`` -- the graph for which modular decomposition tree needs to be
      computed

    - ``g_classes`` -- dictionary (default: ``None``); a dictionary whose values
      are the gamma classes of the graph, and whose keys are a frozenset of the
      vertices corresponding to the class. Used internally.

    OUTPUT: the modular decomposition tree of the graph

    EXAMPLES:

    The Icosahedral graph is Prime::

        sage: from sage.graphs.graph_decompositions.modular_decomposition import *
        sage: print_md_tree(habib_maurer_algorithm(graphs.IcosahedralGraph()))
        PRIME
         3
         4
         7
         9
         11
         1
         5
         8
         0
         2
         6
         10

    The Octahedral graph is not Prime::

        sage: print_md_tree(habib_maurer_algorithm(graphs.OctahedralGraph()))
        SERIES
         PARALLEL
          0
          5
         PARALLEL
          1
          4
         PARALLEL
          2
          3

    Tetrahedral Graph is Series::

        sage: print_md_tree(habib_maurer_algorithm(graphs.TetrahedralGraph()))
        SERIES
         0
         1
         2
         3

    Modular Decomposition tree containing both parallel and series modules::

        sage: d = {2:[4,3,5], 1:[4,3,5], 5:[3,2,1,4], 3:[1,2,5], 4:[1,2,5]}
        sage: g = Graph(d)
        sage: print_md_tree(habib_maurer_algorithm(g))
        SERIES
         PARALLEL
          1
          2
         PARALLEL
          3
          4
         5

    Graph from Marc Tedder implementation of modular decomposition::

        sage: d = {1:[5,4,3,24,6,7,8,9,2,10,11,12,13,14,16,17], 2:[1],
        ....:       3:[24,9,1], 4:[5,24,9,1], 5:[4,24,9,1], 6:[7,8,9,1],
        ....:       7:[6,8,9,1], 8:[6,7,9,1], 9:[6,7,8,5,4,3,1], 10:[1],
        ....:       11:[12,1], 12:[11,1], 13:[14,16,17,1], 14:[13,17,1],
        ....:       16:[13,17,1], 17:[13,14,16,18,1], 18:[17], 24:[5,4,3,1]}
        sage: g = Graph(d)
        sage: test_modular_decomposition(habib_maurer_algorithm(g), g)
        True

    Tetrahedral Graph is Series::

        sage: print_md_tree(habib_maurer_algorithm(graphs.TetrahedralGraph()))
        SERIES
         0
         1
         2
         3

    Modular Decomposition tree containing both parallel and series modules::

        sage: d = {2:[4,3,5], 1:[4,3,5], 5:[3,2,1,4], 3:[1,2,5], 4:[1,2,5]}
        sage: g = Graph(d)
        sage: print_md_tree(habib_maurer_algorithm(g))
        SERIES
         PARALLEL
          1
          2
         PARALLEL
          3
          4
         5

    TESTS:

    Ensure that a random graph and an isomorphic graph have identical modular
    decompositions. ::

        sage: from sage.graphs.graph_decompositions.modular_decomposition import permute_decomposition
        sage: permute_decomposition(2, habib_maurer_algorithm, 20, 0.5)                 # needs sage.groups
    """
    if graph.order() == 1:
        root = Node.create_leaf(next(graph.vertex_iterator()))
        return root

    elif not graph.is_connected():
        root = Node(NodeType.PARALLEL)
        root.children = [habib_maurer_algorithm(graph.subgraph(vertices=sg), g_classes)
                         for sg in graph.connected_components(sort=False)]
        return root

    g_comp = graph.complement()
    if g_comp.is_connected():
        from collections import defaultdict
        root = Node(NodeType.PRIME)
        if g_classes is None:
            g_classes = gamma_classes(graph)
        vertex_set = frozenset(graph)
        edges = [tuple(e) for e in g_classes[vertex_set]]
        sub = graph.subgraph(edges=edges)
        d = defaultdict(list)
        for v in sub:
            for v1 in sub.neighbor_iterator(v):
                d[v1].append(v)
        d1 = defaultdict(list)
        for k, v in d.items():
            d1[frozenset(v)].append(k)
        root.children = [habib_maurer_algorithm(graph.subgraph(vertices=sg), g_classes)
                         for sg in d1.values()]
        return root

    root = Node(NodeType.SERIES)
    root.children = [habib_maurer_algorithm(graph.subgraph(vertices=sg), g_classes)
                     for sg in g_comp.connected_components(sort=False)]
    return root


################################################################################
#                   Exported modular_decomposition function                    #
################################################################################
def modular_decomposition(G, algorithm=None):
    r"""
    Return the modular decomposition of the current graph.

    This function should not be used directly, it should be called via the
    ``modular_decomposition`` method of ``Graph``.

    INPUT:

    - ``G`` -- graph whose modular decomposition tree is to be computed

    - ``algorithm`` -- string (default: ``None``); the algorithm to use among:

      - ``None`` or ``'corneil_habib_paul_tedder'`` -- will use the
        Corneil-Habib-Paul-Tedder algorithm from [TCHP2008]_, its complexity
        is linear in the number of vertices and edges.

      - ``'habib_maurer'`` -- will use the Habib-Maurer algorithm from
        [HM1979]_, its complexity is cubic in the number of vertices.

    OUTPUT: The modular decomposition tree, as an object of type ``Node``.

    TESTS::

        sage: from sage.graphs.graph_decompositions.modular_decomposition import *

        sage: modular_decomposition(Graph())
        EMPTY []

        sage: modular_decomposition(Graph(1))
        NORMAL [0]

        sage: check_algos_are_equivalent(5,                                             # needs networkx
        ....:                   lambda : graphs.RandomProperIntervalGraph(100))

        sage: check_algos_are_equivalent(5, lambda : graphs.RandomGNM(75, 1000))        # needs networkx

        sage: modular_decomposition(DiGraph())
        Traceback (most recent call last):
        ...
        TypeError: the input must be an undirected Sage graph

        sage: modular_decomposition(Graph(5, loops=True))
        Traceback (most recent call last):
        ...
        ValueError: This method is not known to work on graphs with loops...

        sage: modular_decomposition(Graph(5, multiedges=True))
        Traceback (most recent call last):
        ...
        ValueError: This method is not known to work on graphs with multiedges...

        sage: modular_decomposition(Graph(), algorithm='silly walk')
        Traceback (most recent call last):
        ...
        ValueError: unknown algorithm "silly walk"
    """
    from sage.graphs.graph import Graph
    if not isinstance(G, Graph):
        raise TypeError("the input must be an undirected Sage graph")
    G._scream_if_not_simple()

    if algorithm is None:
        algorithm = "corneil_habib_paul_tedder"

    if algorithm not in ("habib_maurer", "corneil_habib_paul_tedder"):
        raise ValueError(f'unknown algorithm "{algorithm}"')

    if not G.order():
        return Node(NodeType.EMPTY)
    if G.order() == 1:
        D = Node(NodeType.NORMAL)
        D.children.append(next(G.vertex_iterator()))
        return D
    if algorithm == "habib_maurer":
        return habib_maurer_algorithm(G)
    return corneil_habib_paul_tedder_algorithm(G)


# ============================================================================
# Below functions are implemented to test the modular decomposition tree
# ============================================================================
# Function implemented for testing
def test_modular_decomposition(tree_root, graph):
    """
    Test the input modular decomposition tree using recursion.

    INPUT:

    - ``tree_root`` -- root of the modular decomposition tree to be tested

    - ``graph`` -- graph whose modular decomposition tree needs to be tested

    OUTPUT: ``True`` if input tree is a modular decomposition else ``False``

    EXAMPLES::

        sage: from sage.graphs.graph_decompositions.modular_decomposition import *
        sage: g = graphs.HexahedralGraph()
        sage: test_modular_decomposition(modular_decomposition(g), g)
        True
    """
    if tree_root.node_type != NodeType.NORMAL:
        for module in tree_root.children:
            if not test_module(module, graph):
                # test whether modules pass the defining
                # characteristics of modules
                return False
            if not test_modular_decomposition(module,
                                              graph.subgraph(get_vertices(module))):
                # recursively test the modular decomposition subtrees
                return False

        if not test_maximal_modules(tree_root, graph):
            # test whether the mdoules are maximal in nature
            return False

    return True


# Function implemented for testing
def test_maximal_modules(tree_root, graph):
    r"""
    Test the maximal nature of modules in a modular decomposition tree.

    Suppose the module `M = [M_1, M_2, \cdots, n]` is the input modular
    decomposition tree. Algorithm forms pairs like `(M_1, M_2), (M_1, M_3),
    \cdots, (M_1, M_n)`; `(M_2, M_3), (M_2, M_4), \cdots, (M_2, M_n)`; `\cdots`
    and so on and tries to form a module using the pair. If the module formed
    has same type as `M` and is of type ``SERIES`` or ``PARALLEL`` then the
    formed module is not considered maximal. Otherwise it is considered maximal
    and `M` is not a modular decomposition tree.

    INPUT:

    - ``tree_root`` -- modular decomposition tree whose modules are tested for
      maximal nature

    - ``graph`` -- graph whose modular decomposition tree is tested

    OUTPUT:

    ``True`` if all modules at first level in the modular decomposition tree
    are maximal in nature

    EXAMPLES::

        sage: from sage.graphs.graph_decompositions.modular_decomposition import *
        sage: g = graphs.HexahedralGraph()
        sage: test_maximal_modules(modular_decomposition(g), g)
        True
    """
    if tree_root.node_type != NodeType.NORMAL:
        for index, module in enumerate(tree_root.children):
            for other_index in range(index + 1, len(tree_root.children)):

                # compute the module formed using modules at index and
                # other_index
                module_formed = form_module(index, other_index,
                                            tree_root, graph)

                if module_formed[0]:
                    # Module formed and the parent of the formed module
                    # should not both be of type SERIES or PARALLEL
                    mod_type = get_module_type(graph.subgraph(module_formed[1]))
                    if (mod_type == tree_root.node_type and
                            (tree_root.node_type == NodeType.PARALLEL or
                             tree_root.node_type == NodeType.SERIES)):
                        continue
                    return False
    return True


def get_vertices(component_root):
    """
    Compute the list of vertices in the (co)component.

    INPUT:

    - ``component_root`` -- root of the (co)component whose vertices need to be
      returned as a list

    OUTPUT:

    list of vertices in the (co)component

    EXAMPLES::

        sage: from sage.graphs.graph_decompositions.modular_decomposition import *
        sage: forest = Node(NodeType.PRIME)
        sage: forest.children = [Node.create_leaf(2), Node.create_leaf(0),
        ....:                    Node.create_leaf(3), Node.create_leaf(1)]
        sage: series_node = Node(NodeType.SERIES)
        sage: series_node.children = [Node.create_leaf(4),
        ....:                         Node.create_leaf(5)]
        sage: parallel_node = Node(NodeType.PARALLEL)
        sage: parallel_node.children = [Node.create_leaf(6),
        ....:                           Node.create_leaf(7)]
        sage: forest.children.insert(1, series_node)
        sage: forest.children.insert(3, parallel_node)
        sage: get_vertices(forest)
        [2, 4, 5, 0, 6, 7, 3, 1]
    """
    vertices = []

    # inner recursive function to recurse over the elements in the
    # ``component``
    def recurse_component(node, vertices):
        if node.node_type == NodeType.NORMAL:
            vertices.append(node.children[0])
            return
        for child in node.children:
            recurse_component(child, vertices)

    recurse_component(component_root, vertices)
    return vertices


# Function implemented for testing
def get_module_type(graph):
    """
    Return the module type of the root of the modular decomposition tree of
    ``graph``.

    INPUT:

    - ``graph`` -- input sage graph

    OUTPUT:

    ``PRIME`` if graph is PRIME, ``PARALLEL`` if graph is PARALLEL and
    ``SERIES`` if graph is of type SERIES

    EXAMPLES::

        sage: from sage.graphs.graph_decompositions.modular_decomposition import get_module_type
        sage: g = graphs.HexahedralGraph()
        sage: get_module_type(g)
        PRIME
    """
    if not graph.is_connected():
        return NodeType.PARALLEL
    elif graph.complement().is_connected():
        return NodeType.PRIME
    return NodeType.SERIES


# Function implemented for testing
def form_module(index, other_index, tree_root, graph):
    r"""
    Forms a module out of the modules in the module pair.

    Let `M_1` and `M_2` be the input modules. Let `V` be the set of vertices in
    these modules. Suppose `x` is a neighbor of subset of the vertices in `V`
    but not all the vertices and `x` does not belong to `V`. Then the set of
    modules also include the module which contains `x`. This process is repeated
    until a module is formed and the formed module if subset of `V` is returned.

    INPUT:

    - ``index`` -- first module in the module pair

    - ``other_index`` -- second module in the module pair

    - ``tree_root`` -- modular decomposition tree which contains the modules
      in the module pair

    - ``graph`` -- graph whose modular decomposition tree is created

    OUTPUT:

    ``[module_formed, vertices]`` where ``module_formed`` is ``True`` if
    module is formed else ``False`` and ``vertices`` is a list of vertices
    included in the formed module

    EXAMPLES::

        sage: from sage.graphs.graph_decompositions.modular_decomposition import *
        sage: g = graphs.HexahedralGraph()
        sage: tree_root = modular_decomposition(g)
        sage: form_module(0, 2, tree_root, g)
        [False, {0, 1, 2, 3, 4, 5, 6, 7}]
    """
    vertices = set(get_vertices(tree_root.children[index]))
    vertices.update(get_vertices(tree_root.children[other_index]))

    # stores all neighbors which are common for all vertices in V
    common_neighbors = set()

    # stores all neighbors of vertices in V which are outside V
    all_neighbors = set()

    while True:
        # remove vertices from all_neighbors and common_neighbors
        all_neighbors.difference_update(vertices)
        common_neighbors.difference_update(vertices)

        for v in vertices:
            # stores the neighbors of v which are outside the set of vertices
            neighbor_list = set(graph.neighbors(v))
            neighbor_list.difference_update(vertices)

            # update all_neighbors and common_neighbors using the
            # neighbor_list
            all_neighbors.update(neighbor_list)
            common_neighbors.intersection_update(neighbor_list)

        if all_neighbors == common_neighbors:  # indicates a module is formed

            # module formed covers the entire graph
            if len(vertices) == graph.order():
                return [False, vertices]

            return [True, vertices]

        # add modules containing uncommon neighbors into the formed module
        for v in (all_neighbors - common_neighbors):
            for index in range(len(tree_root.children)):
                if v in get_vertices(tree_root.children[index]):
                    vertices.update(get_vertices(tree_root.children[index]))
                    break


# Function implemented for testing
def test_module(module, graph):
    """
    Test whether input module is actually a module.

    INPUT:

    - ``module`` -- module which needs to be tested

    - ``graph`` -- input sage graph which contains the module

    OUTPUT: ``True`` if input module is a module by definition else ``False``

    EXAMPLES::

        sage: from sage.graphs.graph_decompositions.modular_decomposition import *
        sage: g = graphs.HexahedralGraph()
        sage: tree_root = modular_decomposition(g)
        sage: test_module(tree_root, g)
        True
        sage: test_module(tree_root.children[0], g)
        True
    """
    # A single vertex is a module
    if module.node_type == NodeType.NORMAL:
        return True

    # vertices contained in module
    vertices_in_module = get_vertices(module)

    # vertices outside module
    vertices_outside = list(set(graph.vertices(sort=False)) - set(vertices_in_module))

    # Nested module with only one child
    if module.node_type != NodeType.NORMAL and len(module.children) == 1:
        return False

    # If children of SERIES module are all SERIES modules
    if module.node_type == NodeType.SERIES:
        if children_node_type(module, NodeType.SERIES):
            return False

    # If children of PARALLEL module are all PARALLEL modules
    if module.node_type == NodeType.PARALLEL:
        if children_node_type(module, NodeType.PARALLEL):
            return False

    # check the module by definition. Vertices in a module should all either
    # be connected or disconnected to any vertex outside module
    for v in vertices_outside:
        if not either_connected_or_not_connected(v, vertices_in_module, graph):
            return False
    return True


# Function implemented for testing
def children_node_type(module, node_type):
    """
    Check whether the node type of the children of ``module`` is ``node_type``.

    INPUT:

    - ``module`` -- module which is tested

    - ``node_type`` -- input node_type

    EXAMPLES::

        sage: from sage.graphs.graph_decompositions.modular_decomposition import *
        sage: g = graphs.OctahedralGraph()
        sage: tree_root = modular_decomposition(g)
        sage: print_md_tree(tree_root)
        SERIES
         PARALLEL
          2
          3
         PARALLEL
          1
          4
         PARALLEL
          0
          5
        sage: children_node_type(tree_root, NodeType.SERIES)
        False
        sage: children_node_type(tree_root, NodeType.PARALLEL)
        True
    """
    return all(node.node_type == node_type for node in module.children)


# Function implemented for testing
def either_connected_or_not_connected(v, vertices_in_module, graph):
    """
    Check whether ``v`` is connected or disconnected to all vertices in the
    module.

    INPUT:

    - ``v`` -- vertex tested

    - ``vertices_in_module`` -- list containing vertices in the module

    - ``graph`` -- graph to which the vertices belong

    EXAMPLES::

        sage: from sage.graphs.graph_decompositions.modular_decomposition import *
        sage: g = graphs.OctahedralGraph()
        sage: print_md_tree(modular_decomposition(g))
        SERIES
         PARALLEL
          2
          3
         PARALLEL
          1
          4
         PARALLEL
          0
          5
        sage: either_connected_or_not_connected(2, [1, 4], g)
        True
        sage: either_connected_or_not_connected(2, [3, 4], g)
        False
    """
    # marks whether vertex v is connected to first vertex in the module
    connected = graph.has_edge(vertices_in_module[0], v)

    # if connected is True then all vertices in module should be connected to
    # v else all should be disconnected
    return all(graph.has_edge(u, v) == connected for u in vertices_in_module)


def tree_to_nested_tuple(root):
    r"""
    Convert a modular decomposition tree to a nested tuple.

    INPUT:

    - ``root`` -- the root of the modular decomposition tree

    OUTPUT:

    A tuple whose first element is the type of the root of the tree and whose
    subsequent nodes are either vertex labels in the case of leaves or tuples
    representing the child subtrees.

    EXAMPLES::

        sage: from sage.graphs.graph_decompositions.modular_decomposition import *
        sage: g = graphs.OctahedralGraph()
        sage: tree_to_nested_tuple(modular_decomposition(g))
        (SERIES, [(PARALLEL, [2, 3]), (PARALLEL, [1, 4]), (PARALLEL, [0, 5])])
    """
    if root.node_type == NodeType.NORMAL:
        return root.children[0]
    else:
        return (root.node_type, [tree_to_nested_tuple(x) for x in root.children])


def nested_tuple_to_tree(nest):
    r"""
    Turn a tuple representing the modular decomposition into a tree.

    INPUT:

    - ``nest`` -- a nested tuple of the form returned by
      :meth:`tree_to_nested_tuple`

    OUTPUT: the root node of a modular decomposition tree

    EXAMPLES::

        sage: from sage.graphs.graph_decompositions.modular_decomposition import *
        sage: tree = (NodeType.SERIES, 1, 2, (NodeType.PARALLEL, 3, 4))
        sage: print_md_tree(nested_tuple_to_tree(tree))
        SERIES
         1
         2
         PARALLEL
          3
          4
    """
    if not isinstance(nest, tuple):
        return Node.create_leaf(nest)

    root = Node(nest[0])
    root.children = [nested_tuple_to_tree(n) for n in nest[1:]]
    return root


def equivalent_trees(root1, root2):
    r"""
    Check that two modular decomposition trees are the same.

    Verify that the structure of the trees is the same. Two leaves are
    equivalent if they represent the same vertex, two internal nodes are
    equivalent if they have the same nodes type and the same number of children
    and there is a matching between the children such that each pair of
    children is a pair of equivalent subtrees.

    EXAMPLES::

        sage: from sage.graphs.graph_decompositions.modular_decomposition import *
        sage: t1 = nested_tuple_to_tree((NodeType.SERIES, 1, 2,
        ....:             (NodeType.PARALLEL, 3, 4)))
        sage: t2 = nested_tuple_to_tree((NodeType.SERIES,
        ....:             (NodeType.PARALLEL, 4, 3), 2, 1))
        sage: equivalent_trees(t1, t2)
        True
    """
    # internal definition
    def node_id(root):
        return (root.node_type, frozenset(get_vertices(root)))

    if root1.node_type != root2.node_type:
        return False

    if len(root1.children) != len(root2.children):
        return False

    if root1.node_type == NodeType.NORMAL:
        return root1.children[0] == root2.children[0]

    child_map = {}
    for node in root2.children:
        child_map[node_id(node)] = node

    for node in root1.children:
        id = node_id(node)
        if id not in child_map:
            return False
        if not equivalent_trees(node, child_map[id]):
            return False

    return True


def relabel_tree(root, perm):
    r"""
    Relabel the leaves of a tree according to a dictionary.

    INPUT:

    - ``root`` -- the root of the tree

    - ``perm`` -- a function, dictionary, list, permutation, or ``None``
      representing the relabeling. See
      :meth:`~sage.graphs.generic_graph.GenericGraph.relabel` for description of
      the permutation input.

    EXAMPLES::

        sage: from sage.graphs.graph_decompositions.modular_decomposition import *
        sage: tuple_tree = (NodeType.SERIES, 1, 2, (NodeType.PARALLEL, 3, 4))
        sage: tree = nested_tuple_to_tree(tuple_tree)
        sage: print_md_tree(relabel_tree(tree, (4,3,2,1)))
        SERIES
         4
         3
         PARALLEL
          2
          1
    """
    # If perm is not a dictionary, we build one !
    if perm is None:

        # vertices() returns a sorted list:
        # this guarantees consistent relabeling
        perm = {v: i for i, v in enumerate(get_vertices(root))}

    elif isinstance(perm, dict):
        from copy import copy
        # If all vertices do not have a new label, the code will touch the
        # dictionary. Let us keep the one we received from the user clean !
        perm = copy(perm)

    elif isinstance(perm, (list, tuple)):
        perm = dict(zip(sorted(get_vertices(root)), perm))

    elif isinstance(perm, PermutationGroupElement):
        n = len(get_vertices(root))
        ddict = {}
        for i in range(1, n):
            ddict[i] = perm(i) % n
        if n > 0:
            ddict[0] = perm(n) % n
        perm = ddict

    elif callable(perm):
        perm = {i: perm(i) for i in get_vertices(root)}

    else:
        raise TypeError("type of perm is not supported for relabeling")

    if root.node_type == NodeType.NORMAL:
        return Node.create_leaf(perm[root.children[0]])
    else:
        new_root = Node(root.node_type)
        new_root.children = [relabel_tree(child, perm) for child in root.children]
        return new_root


# =============================================================================
# Random tests
# =============================================================================

@random_testing
def test_gamma_modules(trials, vertices, prob, verbose=False):
    r"""
    Verify that the vertices of each gamma class of a random graph are modules
    of that graph.

    INPUT:

    - ``trials`` -- the number of trials to run

    - ``vertices`` -- the size of the graph to use

    - ``prob`` -- the probability that any given edge is in the graph
      See :meth:`~sage.graphs.generators.random.RandomGNP` for more details

    - ``verbose`` -- print information on each trial

    EXAMPLES::

        sage: from sage.graphs.graph_decompositions.modular_decomposition import *
        sage: test_gamma_modules(3, 7, 0.5)
    """
    from sage.graphs.generators.random import RandomGNP
    for _ in range(trials):
        g = RandomGNP(vertices, prob)
        if verbose:
            print(g.graph6_string())
        g_classes = gamma_classes(g)
        for module in g_classes.keys():
            m_list = list(module)
            for v in g:
                if v not in module:
                    assert either_connected_or_not_connected(v, m_list, g)
        if verbose:
            print("Passes!")


@random_testing
def permute_decomposition(trials, algorithm, vertices, prob, verbose=False):
    r"""
    Check that a graph and its permuted relabeling have the same modular
    decomposition.

    We generate a ``trials`` random graphs and then generate an isomorphic graph
    by relabeling the original graph. We then verify

    EXAMPLES::

        sage: from sage.graphs.graph_decompositions.modular_decomposition import *
        sage: permute_decomposition(30, habib_maurer_algorithm, 10, 0.5)
    """
    from sage.graphs.generators.random import RandomGNP
    from sage.combinat.permutation import Permutations
    for _ in range(trials):
        g1 = RandomGNP(vertices, prob)
        random_perm = Permutations(list(g1)).random_element()
        g2 = g1.relabel(perm=random_perm, inplace=False)
        if verbose:
            print(g1.graph6_string())
            print(random_perm)
        t1 = algorithm(g1)
        t2 = algorithm(g2)
        assert test_modular_decomposition(t1, g1)
        assert test_modular_decomposition(t2, g2)
        t1p = relabel_tree(t1, random_perm)
        assert equivalent_trees(t1p, t2)
        if verbose:
            print("Passes!")


def random_md_tree(max_depth, max_fan_out, leaf_probability):
    r"""
    Create a random MD tree.

    INPUT:

    - ``max_depth`` -- the maximum depth of the tree

    - ``max_fan_out`` -- the maximum number of children a node can have
      (must be >=4 as a prime node must have at least 4 vertices)

    - ``leaf_probability`` -- the probability that a subtree is a leaf

    EXAMPLES::

        sage: from sage.graphs.graph_decompositions.modular_decomposition import *
        sage: set_random_seed(0)
        sage: tree_to_nested_tuple(random_md_tree(2, 5, 0.5))
        (PRIME, [0, 1, (PRIME, [2, 3, 4, 5, 6]), 7, (PARALLEL, [8, 9, 10])])
    """

    from sage.misc.prandom import choice, randint, random

    if max_fan_out < 4:
        raise ValueError("max_fan_out must be at least 4")

    # Internal function
    def rand_md_tree(max_depth, parent_type):
        r"""
        Create the subtrees of a node.

        A child of a node cannot have the same node type as its parent if its
        parent's node type is either PARALLEL or SERIES.  Also its ``max_depth``
        is one less than its parent's.
        """
        if random() < leaf_probability or max_depth == 1:
            root = Node.create_leaf(current_leaf[0])
            current_leaf[0] += 1
            return root
        if parent_type == NodeType.PRIME:
            node_type = choice([NodeType.PRIME, NodeType.SERIES, NodeType.PARALLEL])
        elif parent_type == NodeType.SERIES:
            node_type = choice([NodeType.PRIME, NodeType.PARALLEL])
        else:
            node_type = choice([NodeType.PRIME, NodeType.SERIES])
        if node_type == NodeType.PRIME:
            num_children = randint(4, max_fan_out)
        else:
            num_children = randint(2, max_fan_out)
        root = Node(node_type)
        root.children = [rand_md_tree(max_depth - 1, node_type)
                         for _ in range(num_children)]
        return root

    # a hack around python2's lack of 'nonlocal'
    current_leaf = [0]
    node_type = choice([NodeType.PRIME, NodeType.SERIES, NodeType.PARALLEL])
    num_children = randint(4, max_fan_out)
    root = Node(node_type)
    root.children = [rand_md_tree(max_depth, node_type)
                     for _ in range(num_children)]
    return root


def md_tree_to_graph(root, prime_node_generator=None):
    r"""
    Create a graph having the given MD tree.

    For the prime nodes, the parameter ``prime_node_generator`` is called with
    the number of vertices as the only argument. If it is ``None``, the path
    graph is used (it is prime when the length is 4 or more).

    EXAMPLES::

        sage: from sage.graphs.graph_decompositions.modular_decomposition import *
        sage: from sage.graphs.graph_generators import graphs
        sage: tup1 = (NodeType.PRIME, 1, (NodeType.SERIES, 2, 3),
        ....:        (NodeType.PARALLEL, 4, 5), 6)
        sage: tree1 = nested_tuple_to_tree(tup1)
        sage: g1 = md_tree_to_graph(tree1)
        sage: g2 = Graph({1: [2, 3], 2: [1, 3, 4, 5], 3: [1, 2, 4, 5],
        ....:             4: [2, 3, 6], 5: [2, 3, 6], 6: [4, 5]})
        sage: g1.is_isomorphic(g2)
        True

        sage: G = md_tree_to_graph(Node(NodeType.EMPTY))
        sage: G.is_isomorphic(Graph())
        True

        sage: tree = Node(NodeType.SERIES)
        sage: tree.children.extend(Node.create_leaf(i) for i in range(5))
        sage: G = md_tree_to_graph(tree)
        sage: G.is_isomorphic(graphs.CompleteGraph(5))
        True

        sage: tree = Node(NodeType.PRIME)
        sage: tree.children.extend(Node.create_leaf(i) for i in range(5))
        sage: png = lambda n: (graphs.PathGraph if n == 4 else graphs.CycleGraph)(n)
        sage: G = md_tree_to_graph(tree, prime_node_generator=png)
        sage: G.is_isomorphic(graphs.CycleGraph(5))
        True
    """
    from itertools import product, combinations
    from sage.graphs.graph import Graph

    if prime_node_generator is None:
        from sage.graphs.graph_generators import graphs
        prime_node_generator = graphs.PathGraph

    def tree_to_vertices_and_edges(root):
        r"""
        Give the list of vertices and edges of the graph having the given md tree.
        """
        if root.is_leaf():
            return (root.children, [])
        children_ve = [tree_to_vertices_and_edges(child) for child in root.children]
        vertices = [v for vs, es in children_ve for v in vs]
        edges = [e for vs, es in children_ve for e in es]
        vertex_lists = [vs for vs, es in children_ve]
        if root.is_prime():
            G = prime_node_generator(len(vertex_lists))
            G.relabel(range(len(vertex_lists)))
            for i1, i2 in G.edge_iterator(labels=False):
                edges.extend(product(vertex_lists[i1], vertex_lists[i2]))
        elif root.is_series():
            for vs1, vs2 in combinations(vertex_lists, 2):
                edges.extend(product(vs1, vs2))
        # else: no edge to be created for PARALLEL nodes
        return (vertices, edges)

    if root.is_empty():
        return Graph()
    vs, es = tree_to_vertices_and_edges(root)
    return Graph([vs, es], format='vertices_and_edges')


@random_testing
def recreate_decomposition(trials, algorithm, max_depth, max_fan_out,
                           leaf_probability, verbose=False):
    r"""
    Verify that we can recreate a random MD tree.

    We create a random MD tree, then create a graph having that decomposition,
    then find a modular decomposition for that graph, and verify that the two
    modular decomposition trees are equivalent.

    EXAMPLES::

        sage: from sage.graphs.graph_decompositions.modular_decomposition import *
        sage: recreate_decomposition(3, habib_maurer_algorithm, 4, 6, 0.5,
        ....:                         verbose=False)
    """
    for _ in range(trials):
        rand_tree = random_md_tree(max_depth, max_fan_out, leaf_probability)
        if verbose:
            print_md_tree(rand_tree)
        graph = md_tree_to_graph(rand_tree)
        if verbose:
            print(graph.graph6_string())
            print(graph.to_dictionary())
        reconstruction = algorithm(graph)
        if verbose:
            print_md_tree(reconstruction)
        assert equivalent_trees(rand_tree, reconstruction)
        if verbose:
            print("Passes!")


@random_testing
def check_algos_are_equivalent(trials, graph_gen, verbose=False):
    r"""
    Verify that both algorithms compute the same tree (up to equivalence) for
    random graphs.

    INPUT:

    - ``trials`` -- integer; the number of tests the function will run.

    - ``graph_gen`` -- function; a function that can be called without argument
      and returns a random graph.

    - ``verbose`` -- boolean (defaul: ``False``); enable printing debug
      information.

    OUTPUT: ``None``. Raises an ``AssertionError`` on failure.

    EXAMPLES::

        sage: from sage.graphs.graph_decompositions.modular_decomposition import *
        sage: check_algos_are_equivalent(3, lambda : graphs.RandomGNP(10, 0.1))
        sage: check_algos_are_equivalent(3, lambda : graphs.RandomGNP(10, 0.5))
        sage: check_algos_are_equivalent(3, lambda : graphs.RandomGNP(10, 0.9))
    """
    for _ in range(trials):
        graph = graph_gen()
        if verbose:
            print(graph.graph6_string())
            print(graph.to_dictionary())
        MD = []
        for algo in ('habib_maurer', 'corneil_habib_paul_tedder'):
            md = modular_decomposition(graph, algorithm=algo)
            MD.append(md)
            if verbose:
                print(f'Using {algo}:')
                print_md_tree(md)
        assert equivalent_trees(MD[0], MD[1])
        if verbose:
            print("Passes!")



# sage_setup: distribution = sagemath-graphs
r"""
Domination

This module implements methods related to the notion of domination in graphs,
and more precisely:

.. csv-table::
    :class: contentstable
    :widths: 30, 70
    :delim: |

    :meth:`~dominating_set` | Return a minimum distance-`k` dominating set of the graph.
    :meth:`~dominating_sets` | Return an iterator over the minimum distance-`k` dominating sets of the graph.
    :meth:`~minimal_dominating_sets` | Return an iterator over the minimal dominating sets of a graph.
    :meth:`~is_dominating` | Check whether a set of vertices dominates a graph.
    :meth:`~is_redundant` | Check whether a set of vertices has redundant vertices (with respect to domination).
    :meth:`~private_neighbors` | Return the private neighbors of a vertex with respect to other vertices.
    :meth:`~greedy_dominating_set` | Return a greedy distance-`k` dominating set of the graph.
    :meth:`~maximum_leaf_number` | Return the maximum leaf number of the graph.

EXAMPLES:

We compute the size of a minimum dominating set of the Petersen graph::

    sage: g = graphs.PetersenGraph()
    sage: g.dominating_set(value_only=True)                                             # needs sage.numerical.mip
    3

We enumerate the minimal dominating sets of the 5-star graph::

    sage: g = graphs.StarGraph(5)
    sage: list(g.minimal_dominating_sets())
    [{0}, {1, 2, 3, 4, 5}]

Now only those that dominate the middle vertex::

    sage: list(g.minimal_dominating_sets([0]))
    [{0}, {1}, {2}, {3}, {4}, {5}]

Now the minimal dominating sets of the 5-path graph::

    sage: g = graphs.PathGraph(5)
    sage: list(g.minimal_dominating_sets())
    [{0, 2, 4}, {1, 4}, {0, 3}, {1, 3}]

We count the minimal dominating sets of the Petersen graph::

    sage: sum(1 for _ in graphs.PetersenGraph().minimal_dominating_sets())
    27

Methods
-------
"""

# ****************************************************************************
#       Copyright (C) 2009 Nathann Cohen <nathann.cohen@gmail.com>
#                     2019 Jean-Florent Raymond <j-florent.raymond@uca.fr>
#                     2023 David Coudert <david.coudert@inria.fr>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from copy import copy
from sage.rings.integer import Integer


def is_dominating(G, dom, focus=None):
    r"""
    Check whether ``dom`` is a dominating set of ``G``.

    We say that a set `D` of vertices of a graph `G` dominates a set `S` if
    every vertex of `S` either belongs to `D` or is adjacent to a vertex of `D`.
    Also, `D` is a dominating set of `G` if it dominates `V(G)`.

    INPUT:

    - ``dom`` -- iterable of vertices of ``G``; the vertices of the supposed
      dominating set

    - ``focus`` -- iterable of vertices of ``G`` (default: ``None``); if
      specified, this method checks instead if ``dom`` dominates the vertices in
      ``focus``

    EXAMPLES::

        sage: g = graphs.CycleGraph(5)
        sage: g.is_dominating([0,1], [4, 2])
        True

        sage: g.is_dominating([0,1])
        False
    """
    to_dom = set(G) if focus is None else set(focus)

    for v in dom:
        if not to_dom:
            return True
        to_dom.difference_update(G.neighbor_iterator(v, closed=True))

    return not to_dom


def is_redundant(G, dom, focus=None):
    r"""
    Check whether ``dom`` has redundant vertices.

    For a graph `G` and sets `D` and `S` of vertices, we say that a vertex `v
    \in D` is *redundant* in `S` if `v` has no private neighbor with respect to
    `D` in `S`.  In other words, there is no vertex in `S` that is dominated by
    `v` but not by `D \setminus \{v\}`.

    INPUT:

    - ``dom`` -- iterable of vertices of ``G``; where we look for redundant
      vertices

    - ``focus`` -- iterable of vertices of ``G`` (default: ``None``); if
      specified, this method checks instead whether ``dom`` has a redundant
      vertex in ``focus``

    .. WARNING::

        The assumption is made that ``focus`` (if provided) does not contain
        repeated vertices.

    EXAMPLES::

        sage: G = graphs.CubeGraph(3)
        sage: G.is_redundant(['000', '101'], ['011'])
        True
        sage: G.is_redundant(['000', '101'])
        False
    """
    dom = list(dom)
    focus = G if focus is None else set(focus)

    # dominator[v] (for v in focus) will be equal to:
    #  - (0, None) if v has no neighbor in dom
    #  - (1, u) if v has a unique neighbor in dom, u
    #  - (2, None) if v has >= 2 neighbors in dom

    # Initialization
    dominator = {v: (0, None) for v in focus}

    for x in dom:
        for v in G.neighbor_iterator(x, closed=True):
            if v in focus:
                # remember about x only if we never encountered
                # neighbors of v so far
                if dominator[v][0] == 0:
                    dominator[v] = (1, x)
                elif dominator[v][0] == 1:
                    dominator[v] = (2, None)

    # We now compute the subset of vertices of dom that have a private neighbor
    # in focus. A vertex v in dom has a private neighbor in focus if it is
    # dominated by a unique vertex, that is if dominator[v][0] == 1
    with_private = set()
    for v in focus:
        if dominator[v][0] == 1:
            with_private.add(dominator[v][1])

    # By construction with_private is a subset of dom and we assume the elements
    # of dom to be unique, so the following is equivalent to checking
    # with_private != set(dom)
    return len(with_private) != len(dom)


def private_neighbors(G, vertex, dom):
    r"""
    Return the private neighbors of a vertex with respect to other vertices.

    A private neighbor of a vertex `v` with respect to a vertex subset `D`
    is a closed neighbor of `v` that is not dominated by a vertex of `D
    \setminus \{v\}`.

    INPUT:

    - ``vertex`` -- a vertex of ``G``

    - ``dom`` -- iterable of vertices of ``G``; the vertices possibly stealing
      private neighbors from ``vertex``

    OUTPUT:

    Return the closed neighbors of ``vertex`` that are not closed neighbors
    of any other vertex of ``dom``.

    EXAMPLES::

        sage: g = graphs.PathGraph(5)
        sage: list(g.private_neighbors(1, [1, 3, 4]))
        [1, 0]

        sage: list(g.private_neighbors(1, [3, 4]))
        [1, 0]

        sage: list(g.private_neighbors(1, [3, 4, 0]))
        []
    """
    # The set of all vertices that are dominated by vertex_subset - vertex:
    closed_neighborhood_vs = set()
    for u in dom:
        if u != vertex:
            closed_neighborhood_vs.update(
                G.neighbor_iterator(u, closed=True))

    return (neighbor
            for neighbor in G.neighbor_iterator(vertex, closed=True)
            if neighbor not in closed_neighborhood_vs)


# ==============================================================================
# Computation of minimum dominating sets
# ==============================================================================

def dominating_sets(g, k=1, independent=False, total=False, connected=False,
                    solver=None, verbose=0, *, integrality_tolerance=1e-3):
    r"""
    Return an iterator over the minimum distance-`k` dominating sets
    of the graph.

    A minimum dominating set `S` of a graph `G` is a set of its vertices of
    minimal cardinality such that any vertex of `G` is in `S` or has one of its
    neighbors in `S`. See the :wikipedia:`Dominating_set`.

    A minimum distance-`k` dominating set is a set `S` of vertices of `G` of
    minimal cardinality such that any vertex of `G` is in `S` or at distance at
    most `k` from a vertex in `S`. A distance-`0` dominating set is the set of
    vertices itself, and when `k` is the radius of the graph, any vertex
    dominates all the other vertices.

    As an optimization problem, it can be expressed as follows, where `N^k(u)`
    denotes the set of vertices at distance at most `k` from `u` (the set of
    neighbors when `k=1`):

    .. MATH::

        \mbox{Minimize : }&\sum_{v\in G} b_v\\
        \mbox{Such that : }&\forall v \in G, b_v+\sum_{u \in N^k(v)} b_u\geq 1\\
        &\forall x\in G, b_x\mbox{ is a binary variable}

    We use constraints generation to iterate over the minimum distance-`k`
    dominating sets. That is, after reporting a solution, we add a constraint to
    discard it and solve the problem again until no more solution can be found.

    INPUT:

    - ``k`` -- nonnegative integer (default: `1`); the domination distance

    - ``independent`` -- boolean (default: ``False``); when ``True``, computes
      minimum independent dominating sets, that is minimum dominating sets that
      are also independent sets (see also
      :meth:`~sage.graphs.graph.independent_set`)

    - ``total`` -- boolean (default: ``False``); when ``True``, computes total
      dominating sets (see the See the :wikipedia:`Dominating_set`)

    - ``connected`` -- boolean (default: ``False``); when ``True``, computes
      connected dominating sets (see :wikipedia:`Connected_dominating_set`)

    - ``solver`` -- string (default: ``None``); specifies a Mixed Integer Linear
      Programming (MILP) solver to be used. If set to ``None``, the default one
      is used. For more information on MILP solvers and which default solver is
      used, see the method :meth:`solve
      <sage.numerical.mip.MixedIntegerLinearProgram.solve>` of the class
      :class:`MixedIntegerLinearProgram
      <sage.numerical.mip.MixedIntegerLinearProgram>`.

    - ``verbose`` -- integer (default: 0); sets the level of verbosity. Set
      to 0 by default, which means quiet.

    - ``integrality_tolerance`` -- float; parameter for use with MILP solvers
      over an inexact base ring; see
      :meth:`MixedIntegerLinearProgram.get_values`.

    EXAMPLES:

    Number of distance-`k` dominating sets of a Path graph of order 10::

        sage: g = graphs.PathGraph(10)
        sage: [sum(1 for _ in g.dominating_sets(k=k)) for k in range(11)]               # needs sage.numerical.mip
        [1, 13, 1, 13, 25, 2, 4, 6, 8, 10, 10]

    If we build a graph from two disjoint stars, then link their centers we will
    find a difference between the cardinality of an independent set and a stable
    independent set::

        sage: g = 2 * graphs.StarGraph(5)
        sage: g.add_edge(0, 6)
        sage: [sum(1 for _ in g.dominating_sets(k=k)) for k in range(11)]               # needs sage.numerical.mip
        [1, 1, 2, 12, 12, 12, 12, 12, 12, 12, 12]

    The total dominating set of the Petersen graph has cardinality 4::

        sage: G = graphs.PetersenGraph()
        sage: G.dominating_set(total=True, value_only=True)                             # needs sage.numerical.mip
        4
        sage: sorted(G.dominating_sets(k=1))                                            # needs sage.numerical.mip
        [[0, 2, 6],
         [0, 3, 9],
         [0, 7, 8],
         [1, 3, 7],
         [1, 4, 5],
         [1, 8, 9],
         [2, 4, 8],
         [2, 5, 9],
         [3, 5, 6],
         [4, 6, 7]]

    Independent distance-`k` dominating sets of a Path graph::

        sage: # needs sage.numerical.mip
        sage: G = graphs.PathGraph(6)
        sage: sorted(G.dominating_sets(k=1, independent=True))
        [[1, 4]]
        sage: sorted(G.dominating_sets(k=2, independent=True))
        [[0, 3], [0, 4], [0, 5], [1, 3], [1, 4], [1, 5], [2, 4], [2, 5]]
        sage: sorted(G.dominating_sets(k=3, independent=True))
        [[2], [3]]

    The dominating set is calculated for both the directed and undirected graphs
    (modification introduced in :issue:`17905`)::

        sage: # needs sage.numerical.mip
        sage: g = digraphs.Path(3)
        sage: g.dominating_set(value_only=True)
        2
        sage: list(g.dominating_sets())
        [[0, 1], [0, 2]]
        sage: list(g.dominating_sets(k=2))
        [[0]]
        sage: g = graphs.PathGraph(3)
        sage: g.dominating_set(value_only=True)
        1
        sage: next(g.dominating_sets())
        [1]

    Minimum connected dominating sets of the Petersen graph::

        sage: G = graphs.PetersenGraph()
        sage: G.dominating_set(total=True, value_only=True)                             # needs sage.numerical.mip
        4
        sage: sorted(G.dominating_sets(k=1, connected=True))                            # needs sage.numerical.mip
        [[0, 1, 2, 6],
         [0, 1, 4, 5],
         [0, 3, 4, 9],
         [0, 5, 7, 8],
         [1, 2, 3, 7],
         [1, 6, 8, 9],
         [2, 3, 4, 8],
         [2, 5, 7, 9],
         [3, 5, 6, 8],
         [4, 6, 7, 9]]

    Subgraph induced by the dominating set is connected::

        sage: G = graphs.PetersenGraph()
        sage: all(G.subgraph(vertices=dom).is_connected()                               # needs sage.numerical.mip
        ....:     for dom in G.dominating_set(k=1, connected=True))
        True

    Minimum distance-k connected dominating sets of the Tietze graph::

        sage: G = graphs.TietzeGraph()
        sage: sorted(G.dominating_sets(k=2, connected=True))                            # needs sage.numerical.mip
        [[0, 9], [1, 0], [2, 3], [4, 3], [5, 6], [7, 6], [8, 0], [10, 3], [11, 6]]
        sage: sorted(G.dominating_sets(k=3, connected=True))                            # needs sage.numerical.mip
        [[0], [1], [2], [3], [4], [5], [6], [7], [8], [9], [10], [11]]

    TESTS::

        sage: g = Graph([(0, 1)])
        sage: next(g.dominating_sets(k=-1))
        Traceback (most recent call last):
        ...
        ValueError: the domination distance must be a nonnegative integer

    The method is robust to vertices with incomparable labels::

        sage: G = Graph([(1, 'A'), ('A', 2), (2, 3), (3, 1)])
        sage: L = list(G.dominating_sets())                                             # needs sage.numerical.mip
        sage: len(L)                                                                    # needs sage.numerical.mip
        6
    """
    g._scream_if_not_simple(allow_multiple_edges=True, allow_loops=not total)

    if not k:
        yield list(g)
        return
    if k < 0:
        raise ValueError("the domination distance must be a nonnegative integer")

    from sage.numerical.mip import MixedIntegerLinearProgram
    from sage.numerical.mip import MIPSolverException
    p = MixedIntegerLinearProgram(maximization=False, solver=solver,
                                  constraint_generation=True)
    b = p.new_variable(binary=True)

    if k == 1:
        # For any vertex v, one of its neighbors or v itself is in the minimum
        # dominating set. If g is directed, we use the in-neighbors of v
        # instead.
        neighbors_iter = g.neighbor_in_iterator if g.is_directed() else g.neighbor_iterator
    else:
        # When k > 1, we use BFS to determine the vertices that can reach v
        # through a path of length at most k
        gg = g.reverse() if g.is_directed() else g

        def neighbors_iter(x):
            it = gg.breadth_first_search(x, distance=k)
            _ = next(it)
            yield from it

    if total:
        # We want a total dominating set
        for v in g:
            p.add_constraint(p.sum(b[u] for u in neighbors_iter(v)), min=1)
    else:
        for v in g:
            p.add_constraint(b[v] + p.sum(b[u] for u in neighbors_iter(v)), min=1)

    if independent:
        # no two adjacent vertices are in the set
        for u, v in g.edge_iterator(labels=None):
            p.add_constraint(b[u] + b[v], max=1)

    if connected:
        E = set(frozenset(e) for e in g.edge_iterator(labels=False))
        # edges used in the spanning tree
        edge = p.new_variable(binary=True, name='e')
        # relaxed edges to test for acyclicity
        r_edge = p.new_variable(nonnegative=True, name='re')

        # 1. We want a tree
        p.add_constraint(p.sum(edge[fe] for fe in E)
                         == p.sum(b[u] for u in g) - 1)

        # 2. An edge can be in the tree if its end vertices are selected
        for fe in E:
            u, v = fe
            p.add_constraint(edge[fe] <= b[u])
            p.add_constraint(edge[fe] <= b[v])

        # 3. Subtour elimination constraints
        for fe in E:
            u, v = fe
            p.add_constraint(edge[fe] <= r_edge[u, v] + r_edge[v, u])

        eps = 1 / (5 * Integer(g.order()))
        for v in g:
            p.add_constraint(p.sum(r_edge[u, v] for u in g.neighbor_iterator(v)), max=1 - eps)

    # Minimizes the number of vertices used
    p.set_objective(p.sum(b[v] for v in g))

    best = g.order()
    while True:
        try:
            p.solve(log=verbose)
        except MIPSolverException:
            # No more solutions
            break
        b_val = p.get_values(b, convert=bool, tolerance=integrality_tolerance)
        dom = [v for v in g if b_val[v]]
        if len(dom) > best:
            # All minimum solution have been reported
            break
        yield dom
        best = len(dom)
        # Prevent finding twice a solution
        p.add_constraint(p.sum(b[u] for u in dom) <= best - 1)


def dominating_set(g, k=1, independent=False, total=False, connected=False, value_only=False,
                   solver=None, verbose=0, *, integrality_tolerance=1e-3):
    r"""
    Return a minimum distance-`k` dominating set of the graph.

    A minimum dominating set `S` of a graph `G` is a set of its vertices of
    minimal cardinality such that any vertex of `G` is in `S` or has one of its
    neighbors in `S`. See the :wikipedia:`Dominating_set`.

    A minimum distance-`k` dominating set is a set `S` of vertices of `G` of
    minimal cardinality such that any vertex of `G` is in `S` or at distance at
    most `k` from a vertex in `S`. A distance-`0` dominating set is the set of
    vertices itself, and when `k` is the radius of the graph, any vertex
    dominates all the other vertices.

    As an optimization problem, it can be expressed as follows, where `N^k(u)`
    denotes the set of vertices at distance at most `k` from `u` (the set of
    neighbors when `k=1`):

    .. MATH::

        \mbox{Minimize : }&\sum_{v\in G} b_v\\
        \mbox{Such that : }&\forall v \in G, b_v+\sum_{u \in N^k(v)} b_u\geq 1\\
        &\forall x\in G, b_x\mbox{ is a binary variable}

    INPUT:

    - ``k`` -- nonnegative integer (default: `1`); the domination distance

    - ``independent`` -- boolean (default: ``False``); when ``True``, computes a
      minimum independent dominating set, that is a minimum dominating set that
      is also an independent set (see also
      :meth:`~sage.graphs.graph.independent_set`)

    - ``total`` -- boolean (default: ``False``); when ``True``, computes a total
      dominating set (see the See the :wikipedia:`Dominating_set`)

    - ``connected`` -- boolean (default: ``False``); when ``True``, computes a
      connected dominating set (see :wikipedia:`Connected_dominating_set`)

    - ``value_only`` -- boolean (default: ``False``); whether to only return the
      cardinality of the computed dominating set, or to return its list of
      vertices (default)

    - ``solver`` -- string (default: ``None``); specifies a Mixed Integer Linear
      Programming (MILP) solver to be used. If set to ``None``, the default one
      is used. For more information on MILP solvers and which default solver is
      used, see the method :meth:`solve
      <sage.numerical.mip.MixedIntegerLinearProgram.solve>` of the class
      :class:`MixedIntegerLinearProgram
      <sage.numerical.mip.MixedIntegerLinearProgram>`.

    - ``verbose`` -- integer (default: 0); sets the level of verbosity. Set
      to 0 by default, which means quiet.

    - ``integrality_tolerance`` -- float; parameter for use with MILP solvers
      over an inexact base ring; see
      :meth:`MixedIntegerLinearProgram.get_values`.

    EXAMPLES:

    A basic illustration on a ``PappusGraph``::

        sage: g = graphs.PappusGraph()
        sage: g.dominating_set(value_only=True)                                         # needs sage.numerical.mip
        5

    If we build a graph from two disjoint stars, then link their centers we will
    find a difference between the cardinality of an independent set and a stable
    independent set::

        sage: g = 2 * graphs.StarGraph(5)
        sage: g.add_edge(0, 6)
        sage: len(g.dominating_set())                                                   # needs sage.numerical.mip
        2
        sage: len(g.dominating_set(independent=True))                                   # needs sage.numerical.mip
        6

    The total dominating set of the Petersen graph has cardinality 4::

        sage: G = graphs.PetersenGraph()
        sage: G.dominating_set(total=True, value_only=True)                             # needs sage.numerical.mip
        4

    The dominating set is calculated for both the directed and undirected graphs
    (modification introduced in :issue:`17905`)::

        sage: g = digraphs.Path(3)
        sage: g.dominating_set(value_only=True)                                         # needs sage.numerical.mip
        2
        sage: g = graphs.PathGraph(3)
        sage: g.dominating_set(value_only=True)                                         # needs sage.numerical.mip
        1

    Cardinality of distance-`k` dominating sets::

        sage: G = graphs.PetersenGraph()
        sage: [G.dominating_set(k=k, value_only=True) for k in range(G.radius() + 1)]   # needs sage.numerical.mip
        [10, 3, 1]
        sage: G = graphs.PathGraph(5)
        sage: [G.dominating_set(k=k, value_only=True) for k in range(G.radius() + 1)]   # needs sage.numerical.mip
        [5, 2, 1]
    """
    dom = next(dominating_sets(g, k=k, independent=independent, total=total,
                               connected=connected, solver=solver, verbose=verbose,
                               integrality_tolerance=integrality_tolerance))
    return Integer(len(dom)) if value_only else dom

# ==============================================================================
# Enumeration of minimal dominating set as described in [BDHPR2019]_
# ==============================================================================


def _parent(G, dom, V_prev):
    r"""
    Return a subset of dom that is irredundant in ``V_prev``.

    For internal use.

    INPUT:

    - ``G`` -- a graph

    - ``dom`` -- an iterable of vertices of ``G``

    - ``V_prev`` -- an iterable of vertices of ``G``

    OUTPUT:

    Return the list obtained from ``dom`` by iteratively removing those
    vertices of minimum index that have no private neighbor in ``V_prev``.

    TESTS::

        sage: from sage.graphs.domination import _parent
        sage: G = graphs.PathGraph(4)
        sage: G.add_vertices([4, 5])
        sage: G.add_edges([(4, 1), (5, 2)])
        sage: _parent(G, [0, 2, 4, 5], [1, 2])
        [4, 5]
        sage: _parent(G, [0, 2, 4, 5], [1, 3])
        [2]
    """
    # The list where we search vertices
    D_start = sorted(dom, reverse=True)

    # The list to be output at the end, that we construct:
    D_end = []

    while D_start:
        v = D_start.pop()  # element of min index
        priv = set(G.neighbor_iterator(v, closed=True))
        # We remove the vertices already dominated
        # by other vertices of (D_end union D_start)
        priv.difference_update(*(G.neighbor_iterator(u, closed=True)
                                 for u in D_start if u != v))
        priv.difference_update(*(G.neighbor_iterator(u, closed=True)
                                 for u in D_end if u != v))
        # Now priv is the private neighborhood of v in G wrt D_start + D_end
        if priv.intersection(V_prev) != set():
            # if v has a private in V_prev, we keep it
            D_end.append(v)

    return D_end


def _peel(G, A):
    r"""
    Return a peeling of a vertex iterable of a graph.

    For internal use.
    Given a graph `G` and a subset `A` of its vertices, a peeling of `(G,A)` is
    a list `[(u_0, V_0), \dots, (u_p, V_p)]` such that `u_0` is ``None``,
    `V_0` is the empty set, `V_p = A` and for every `i \in \{1, \dots, p\}`,
    `V_{i-1} = V_i \setminus N[v_i]`, for some vertex `u_i` of `V_i`.

    INPUT:

    - ``G`` -- a graph

    - ``A`` -- set of vertices of `G`

    OUTPUT:

    A peeling of `(G, A)`.

    TESTS::

        sage: from sage.graphs.domination import _peel
        sage: G = Graph(10); _peel(G, {0, 1, 2, 3, 4})
        [(None, set()),
        (4, {4}),
        (3, {3, 4}),
        (2, {2, 3, 4}),
        (1, {1, 2, 3, 4}),
        (0, {0, 1, 2, 3, 4})]


        sage: from sage.graphs.domination import _peel
        sage: G = graphs.PathGraph(10); _peel(G, set((i for i in range(10) if i%2==0)))
        [(None, set()),
        (8, {8}),
        (6, {6, 8}),
        (4, {4, 6, 8}),
        (2, {2, 4, 6, 8}),
        (0, {0, 2, 4, 6, 8})]
    """
    Acomp = set(G)
    Acomp.difference_update(A)  # Acomp  = V - A

    peeling = []
    H = copy(G)
    H.delete_vertices(list(Acomp))
    del Acomp

    while H:
        ui = next(H.vertex_iterator())  # pick some vertex of H
        Vi = set(H)
        peeling.append((ui, Vi))
        H.delete_vertices(H.neighbor_iterator(ui, closed=True))
    peeling.append((None, set()))
    peeling.reverse()
    return peeling


def _cand_ext_enum(G, to_dom, u_next):
    r"""
    Return the minimal dominating sets of ``to_dom``.

    For internal use.
    Assumption: ``u_next`` dominates ``to_dom``.

    INPUT:

    - ``G`` -- a graph

    - ``to_dom`` -- a ``set()`` of vertices of ``G``

    - ``u_next`` -- a vertex of ``G`` that dominates ``to_dom``

    OUTPUT: an iterator over the minimal dominating sets of ``to_dom``

    TESTS::

        sage: from sage.graphs.domination import _cand_ext_enum
        sage: g = graphs.DiamondGraph()
        sage: l = list(_cand_ext_enum(g, {0, 2, 3}, 1,))
        sage: len(l) == 3 and {1} in l and {2} in l and {0, 3} in l
        True
    """

    def _aux_with_rep(H, to_dom, u_next):
        """
        Return the minimal dominating sets of ``to_dom``, with the
        assumption that ``u_next`` dominates ``to_som``.

        .. WARNING::

            The same output may be output several times (up to `|H|` times).

        In order to later remove duplicates, we here output pairs ``(ext, i)``
        where ``ext`` is the output candidate extension and ``i`` counts how
        many elements have already been output.
        """
        if u_next not in to_dom:
            # In this case, enumerating the minimal DSs of the subset
            # to_dom is a smaller instance as it excludes u_next:

            cand_ext_index = 0

            for ext in H.minimal_dominating_sets(to_dom):
                yield (ext, cand_ext_index)
                cand_ext_index += 1

        elif to_dom == {u_next}:
            # In this case, only u_next has to be dominated
            cand_ext_index = 0
            for w in H.neighbor_iterator(u_next, closed=True):
                # Notice that the case w = u_next is included
                yield ({w}, cand_ext_index)
                cand_ext_index += 1

        else:
            # In this case, both u_next and to_dom-u_next have to be dominated

            # We first output the trivial output
            # (as to_dom is subset of N(u_next)):
            yield ({u_next}, 0)
            # Start from 1 because we already output the 0-th elt:
            cand_ext_index = 1

            # When u_next is not in the DS, one of its neighbors w should be:
            for w in H.neighbor_iterator(u_next):

                remains_to_dom = set(to_dom)
                remains_to_dom.difference_update(
                    H.neighbor_iterator(w, closed=True))
                # Here again we recurse on a smaller instance at it
                # excludes u_next (and w)
                for Q in H.minimal_dominating_sets(remains_to_dom):
                    ext = set(Q)
                    ext.add(w)
                    # By construction w dominates u_next and Q dominates
                    # to_dom - N[w], so ext dominates to_dom: it is a
                    # valid output iff it is not redundant
                    if not H.is_redundant(ext):
                        yield (ext, cand_ext_index)
                        cand_ext_index += 1
    #
    # End of aux_with_rep routine

    # Here we use aux_with_rep twice to enumerate the minimal
    # dominating sets while avoiding repeated outputs
    for X, i in _aux_with_rep(G, to_dom, u_next):
        for Y, j in _aux_with_rep(G, to_dom, u_next):
            if j >= i:
                # This is the first time we meet X: we output it
                yield X
                break
            elif Y == X:  # These are sets
                # X has already been output in the past: we ignore it
                break


def minimal_dominating_sets(G, to_dominate=None, work_on_copy=True, k=1):
    r"""
    Return an iterator over the minimal dominating sets of a graph.

    INPUT:

    - ``G`` -- a graph

    - ``to_dominate`` -- vertex iterable or ``None`` (default: ``None``);
      the set of vertices to be dominated

    - ``work_on_copy`` -- boolean (default: ``True``); whether or not to work on
      a copy of the input graph; if set to ``False``, the input graph will be
      modified (relabeled)

    - ``k`` -- nonnegative integer (default: `1`); the domination distance

    OUTPUT:

    An iterator over the inclusion-minimal sets of vertices of ``G``.
    If ``to_dominate`` is provided, return an iterator over the
    inclusion-minimal sets of vertices that dominate the vertices of
    ``to_dominate``.

    ALGORITHM: The algorithm described in [BDHPR2019]_.

    AUTHOR: Jean-Florent Raymond (2019-03-04) -- initial version.

    EXAMPLES::

        sage: G = graphs.ButterflyGraph()
        sage: ll = list(G.minimal_dominating_sets())
        sage: pp = [{0, 1}, {1, 3}, {0, 2}, {2, 3}, {4}]
        sage: len(ll) == len(pp) and all(x in pp for x in ll) and all(x in ll for x in pp)
        True

        sage: ll = list(G.minimal_dominating_sets([0,3]))
        sage: pp = [{0}, {3}, {4}]
        sage: len(ll) == len(pp) and all(x in pp for x in ll) and all(x in ll for x in pp)
        True

        sage: ll = list(G.minimal_dominating_sets([4]))
        sage: pp = [{4}, {0}, {1}, {2}, {3}]
        sage: len(ll) == len(pp) and all(x in pp for x in ll) and all(x in ll for x in pp)
        True

    ::

        sage: ll = list(graphs.PetersenGraph().minimal_dominating_sets())
        sage: pp = [{0, 2, 6},
        ....:       {0, 9, 3},
        ....:       {0, 8, 7},
        ....:       {1, 3, 7},
        ....:       {1, 4, 5},
        ....:       {8, 1, 9},
        ....:       {8, 2, 4},
        ....:       {9, 2, 5},
        ....:       {3, 5, 6},
        ....:       {4, 6, 7},
        ....:       {0, 8, 2, 9},
        ....:       {0, 3, 6, 7},
        ....:       {1, 3, 5, 9},
        ....:       {8, 1, 4, 7},
        ....:       {2, 4, 5, 6},
        ....:       {0, 1, 2, 3, 4},
        ....:       {0, 1, 2, 5, 7},
        ....:       {0, 1, 4, 6, 9},
        ....:       {0, 1, 5, 6, 8},
        ....:       {0, 8, 3, 4, 5},
        ....:       {0, 9, 4, 5, 7},
        ....:       {8, 1, 2, 3, 6},
        ....:       {1, 2, 9, 6, 7},
        ....:       {9, 2, 3, 4, 7},
        ....:       {8, 2, 3, 5, 7},
        ....:       {8, 9, 3, 4, 6},
        ....:       {8, 9, 5, 6, 7}]
        sage: len(ll) == len(pp) and all(x in pp for x in ll) and all(x in ll for x in pp)
        True

    Listing minimal distance-`k` dominating sets::

        sage: G = graphs.Grid2dGraph(2, 3)
        sage: list(G.minimal_dominating_sets(k=0))
        [{(0, 0), (0, 1), (0, 2), (1, 0), (1, 1), (1, 2)}]
        sage: list(G.minimal_dominating_sets(k=1))
        [{(0, 0), (0, 2), (1, 1)},
         {(0, 1), (1, 1)},
         {(0, 0), (0, 1), (0, 2)},
         {(0, 2), (1, 0)},
         {(0, 0), (1, 2)},
         {(0, 1), (1, 0), (1, 2)},
         {(1, 0), (1, 1), (1, 2)}]
        sage: list(G.minimal_dominating_sets(k=2))
        [{(0, 0), (1, 2)},
         {(0, 2), (1, 2)},
         {(1, 0), (1, 2)},
         {(0, 1)},
         {(0, 0), (0, 2)},
         {(0, 2), (1, 0)},
         {(0, 0), (1, 0)},
         {(1, 1)}]
        sage: list(G.minimal_dominating_sets(k=3))
        [{(0, 0)}, {(0, 1)}, {(0, 2)}, {(1, 0)}, {(1, 1)}, {(1, 2)}]

    When parameter ``work_on_copy`` is ``False``, the input graph is modified
    (relabeled)::

        sage: G = Graph([('A', 'B')])
        sage: _ = list(G.minimal_dominating_sets(work_on_copy=True))
        sage: set(G) == {'A', 'B'}
        True
        sage: _ = list(G.minimal_dominating_sets(work_on_copy=False))
        sage: set(G) == {'A', 'B'}
        False
        sage: set(G) == {0, 1}
        True

    TESTS:

    The empty graph is handled correctly::

        sage: list(Graph().minimal_dominating_sets())
        [set()]

    Test on all graphs on 6 vertices::

        sage: from sage.combinat.subset import Subsets
        sage: def minimal_dominating_sets_naive(G):
        ....:     return (S for S in Subsets(G.vertices(sort=False))
        ....:             if not(G.is_redundant(S)) and G.is_dominating(S))
        sage: def big_check(n):
        ....:     for G in graphs(n):
        ....:         ll = list(G.minimal_dominating_sets())
        ....:         pp = list(minimal_dominating_sets_naive(G))
        ....:         if (len(pp) != len(pp)
        ....:             or any(x not in pp for x in ll)
        ....:             or any(x not in ll for x in pp)):
        ....:             return False
        ....:     return True
        sage: big_check(6)  # long time
        True

    Outputs are unique::

        sage: def check_uniqueness(g):
        ....:     counter_1 = 0
        ....:     for dom_1 in g.minimal_dominating_sets():
        ....:         counter_1 += 1
        ....:         counter_2 = 0
        ....:         for dom_2 in g.minimal_dominating_sets():
        ....:             counter_2 += 1
        ....:             if counter_2 >= counter_1:
        ....:                 break
        ....:             if dom_1 == dom_2:
        ....:                 return False
        ....:     return True
        sage: check_uniqueness(graphs.RandomGNP(9, 0.5))
        True

    Asking for a negative distance::

        sage: next(Graph(1).minimal_dominating_sets(k=-1))
        Traceback (most recent call last):
        ...
        ValueError: the domination distance must be a nonnegative integer

    Trying to dominate vertices that are not part of the graph::

        sage: next(Graph(1).minimal_dominating_sets(to_dominate=['foo']))
        Traceback (most recent call last):
        ...
        ValueError: vertex (foo) is not a vertex of the graph

    The method is robust to vertices with incomparable labels::

        sage: G = Graph([(1, 'A'), ('A', 2), (2, 3), (3, 1)])
        sage: L = list(G.minimal_dominating_sets())
        sage: len(L)
        6
        sage: {3, 'A'} in L
        True
    """
    def tree_search(H, plng, dom, i):
        r"""
        Enumerate minimal dominating sets recursively.

        INPUT:

        - ``H`` -- a graph

        - ``plng`` -- a peeling of H (result of :func:`_peel`)

        - ``dom`` -- a minimal dominating set of ``plng[i][1]``

        - ``i`` -- integer; the current position in ``plng``

        OUTPUT:

        An iterator over those minimal dominating sets (in ``H``) of
        ``plng[-1][1]`` that are children of ``dom`` (with respect to the
        :func:`parent` function).

        ALGORITHM:

        We iterate over those minimal dominating sets of ``plng[i + 1][1]`` that
        are children of dom and call recursively on each. The fact that we
        iterate over children (with respect to the `parent` function) ensures
        that we do not have repeated outputs.
        """
        if i == len(plng) - 1:
            # we reached a leaf, i.e. dom is a minimal dominating set
            # of plng[i][1] = plng[-1][1]
            yield dom
            return

        u_next, V_next = plng[i + 1]

        if H.is_dominating(dom, V_next):
            # if dom dominates V_next
            # then dom is its unique extension: we recurse on it
            for Di in tree_search(H, plng, dom, i + 1):
                yield Di
            return

        # Otherwise, V_next - <what dom dominates> is what we have to dominate
        to_dom = V_next - set().union(
            *(G.neighbor_iterator(vert, closed=True)
              for vert in dom))

        for can_ext in _cand_ext_enum(H, to_dom, u_next):

            # We complete dom with can_ext -> canD
            canD = set().union(can_ext, dom)

            if (not H.is_redundant(canD, V_next)
                    and set(dom) == set(_parent(H, canD, plng[i][1]))):
                # By construction, can_ext is a dominating set of
                # `V_next - N[dom]`, so canD dominates V_next.
                # If canD is a legitimate child of dom and is not redundant, we
                # recurse on it:
                for Di in tree_search(H, plng, canD, i + 1):
                    yield Di
    ##
    # end of tree-search routine

    if k < 0:
        raise ValueError("the domination distance must be a nonnegative integer")
    if not k:
        yield set(G) if to_dominate is None else set(to_dominate)
        return

    int_to_vertex = list(G)
    vertex_to_int = {u: i for i, u in enumerate(int_to_vertex)}

    if to_dominate is None:
        vertices_to_dominate = set(range(G.order()))
    else:
        for u in to_dominate:
            if u not in G:
                raise ValueError(f"vertex ({u}) is not a vertex of the graph")
        vertices_to_dominate = {vertex_to_int[u] for u in to_dominate}

    if not vertices_to_dominate:
        # base case: vertices_to_dominate is empty
        # the empty set/list is the only minimal DS of the empty set
        yield set()
        return
    if k > 1:
        # We build a graph H with an edge between u and v if these vertices are
        # at distance at most k in G
        H = G.__class__(G.order())
        for u, ui in vertex_to_int.items():
            H.add_edges((ui, vertex_to_int[v])
                        for v in G.breadth_first_search(u, distance=k) if u != v)
        G = H
    elif work_on_copy:
        G = G.relabel(perm=vertex_to_int, inplace=False)
    else:
        # The input graph is modified
        G.relabel(perm=vertex_to_int, inplace=True)

    peeling = _peel(G, vertices_to_dominate)

    for dom in tree_search(G, peeling, set(), 0):
        yield {int_to_vertex[v] for v in dom}


# ==============================================================================
# Greedy heuristic for dominating set
# ==============================================================================

def greedy_dominating_set(G, k=1, vertices=None, ordering=None, return_sets=False, closest=False):
    r"""
    Return a greedy distance-`k` dominating set of the graph.

    A distance-`k` dominating set `S` of a graph `G` is a set of its vertices of
    minimal cardinality such that any vertex of `G` is in `S` or is at distance
    at most `k` from a vertex in `S`. See the :wikipedia:`Dominating_set`.

    When `G` is directed, vertex `u` can be a dominator of vertex `v` if there
    is a directed path of length at most `k` from `u` to `v`.

    This method implements a greedy heuristic to find a minimal dominatic set.

    INPUT:

    - ``G`` -- a Graph

    - ``k`` -- integer (default: `1`); the domination distance to consider

    - ``vertices`` -- iterable container of vertices (default: ``None``); when
      specified, return a dominating set of the specified vertices only

    - ``ordering`` -- string (default: ``None``); specify the order in which to
      consider the vertices

      - ``None`` -- if ``vertices`` is ``None``, then consider the vertices in
        the order given by ``list(G)``. Otherwise, consider the vertices in the
        order of iteration of ``vertices``.

      - ``'degree_min'`` -- consider the vertices by increasing degree

      - ``'degree_max'`` -- consider the vertices by decreasing degree

    - ``return_sets`` -- boolean (default: ``False``); whether to return the
      vertices of the dominating set only (default), or a dictionary mapping
      each vertex of the dominating set to the set of vertices it dominates.

    - ``closest`` -- boolean (default: ``False``); whether to attach a vertex to
      its closest dominator or not. This parameter is use only when
      ``return_sets`` is ``True``.

    EXAMPLES:

    Dominating sets of a path::

        sage: from sage.graphs.domination import greedy_dominating_set
        sage: G = graphs.PathGraph(5)
        sage: sorted(greedy_dominating_set(G, ordering=None))
        [0, 2, 4]
        sage: sorted(greedy_dominating_set(G, ordering='degree_min'))
        [0, 2, 4]
        sage: sorted(greedy_dominating_set(G, ordering='degree_max'))
        [1, 3]
        sage: sorted(greedy_dominating_set(G, k=2, ordering=None))
        [0, 3]
        sage: sorted(greedy_dominating_set(G, k=2, ordering='degree_min'))
        [0, 4]
        sage: sorted(greedy_dominating_set(G, k=2, ordering='degree_max'))
        [1, 4]
        sage: greedy_dominating_set(G, k=3, ordering='degree_min', return_sets=True, closest=False)
        {0: {0, 1, 2, 3}, 4: {4}}
        sage: greedy_dominating_set(G, k=3, ordering='degree_min', return_sets=True, closest=True)
        {0: {0, 2, 3}, 4: {1, 4}}

    Asking for a dominating set of a subset of vertices::

        sage: from sage.graphs.domination import greedy_dominating_set
        sage: from sage.graphs.domination import is_dominating
        sage: G = graphs.PetersenGraph()
        sage: vertices = {0, 1, 2, 3, 4, 5}
        sage: dom = greedy_dominating_set(G, vertices=vertices, return_sets=True)
        sage: sorted(dom)
        [0, 2]
        sage: is_dominating(G, dom, focus=vertices)
        True
        sage: is_dominating(G, dom)
        False
        sage: dominated = [u for v in dom for u in dom[v]]
        sage: sorted(dominated) == sorted(vertices)
        True

    Influence of the ordering of the vertices on the result::

        sage: from sage.graphs.domination import greedy_dominating_set
        sage: G = graphs.StarGraph(4)
        sage: greedy_dominating_set(G, vertices=[0, 1, 2, 3, 4])
        [0]
        sage: sorted(greedy_dominating_set(G, vertices=[1, 2, 3, 4, 0]))
        [1, 2, 3, 4]

    Dominating set of a directed graph::

        sage: from sage.graphs.domination import greedy_dominating_set
        sage: D = digraphs.Path(3)
        sage: sorted(greedy_dominating_set(D, vertices=[0, 1, 2]))
        [0, 2]

    TESTS:

    Random tests::

        sage: from sage.graphs.domination import greedy_dominating_set
        sage: from sage.graphs.domination import is_dominating
        sage: G = graphs.RandomGNP(15, .2)
        sage: for o in [None, "degree_min", "degree_max"]:
        ....:     for c in [True, False]:
        ....:         dom = greedy_dominating_set(G, ordering=o, closest=c)
        ....:         if not is_dominating(G, dom):
        ....:             print("something goes wrong")

    Corner cases::

        sage: greedy_dominating_set(Graph())
        []
        sage: greedy_dominating_set(Graph(1))
        [0]
        sage: greedy_dominating_set(Graph(2))
        [0, 1]
        sage: G = graphs.PathGraph(5)
        sage: dom = greedy_dominating_set(G, vertices=[0, 1, 3, 4])

    The method is robust to vertices with incomparable labels::

        sage: G = Graph([(1, 'A')])
        sage: len(greedy_dominating_set(G))
        1

    Check parameters::

        sage: greedy_dominating_set(G, ordering='foo')
        Traceback (most recent call last):
        ...
        ValueError: ordering must be None, "degree_min" or "degree_max"
    """
    if vertices is None:
        vertices = list(G)
    else:
        vertices = [u for u in vertices if u in G]

    if ordering in ["degree_min", "degree_max"]:
        vertices = sorted(vertices, key=G.degree, reverse=ordering.endswith("max"))
    elif ordering is not None:
        raise ValueError('ordering must be None, "degree_min" or "degree_max"')

    if not G:
        return dict() if return_sets else []
    if not k:
        return vertices

    n = G.order()
    dom = dict()
    seen = set()
    # We want to dominate only the set vertices
    to_avoid = set() if len(vertices) == n else set(G).difference(vertices)

    if closest:
        # Attach each dominated vertex to its closest dominator
        from sage.rings.infinity import Infinity
        dominator = {u: (u, +Infinity) for u in vertices}
        for u in vertices:
            if u in seen:
                continue
            dom[u] = set()
            for v, d in G.breadth_first_search(u, distance=k, report_distance=True):
                if v in to_avoid:
                    continue
                if v not in seen:
                    dom[u].add(v)
                    seen.add(v)
                    dominator[v] = (u, d)
                else:
                    x, dx = dominator[v]
                    if dx < d:
                        dom[x].discard(v)
                        dom[u].add(v)
                        dominator[v] = (u, d)

    else:
        for u in vertices:
            if u in seen:
                continue
            dom[u] = set()
            for v in G.breadth_first_search(u, distance=k):
                if v not in to_avoid and v not in seen:
                    dom[u].add(v)
                    seen.add(v)

    if return_sets:
        return dom
    else:
        return list(dom)


def maximum_leaf_number(G, solver=None, verbose=0, integrality_tolerance=1e-3):
    r"""
    Return the maximum leaf number of the graph.

    The maximum leaf number is the maximum possible number of leaves of a
    spanning tree of `G`. This is also the cardinality of the complement of a
    minimum connected dominating set.
    See the :wikipedia:`Connected_dominating_set`.

    The MLN of a graph with less than 2 vertices is 0, while the MLN of a connected
    graph with 2 or 3 vertices is 1 or 2 respectively.

    INPUT:

    - ``G`` -- a Graph

    - ``solver`` -- string (default: ``None``); specifies a Mixed Integer Linear
      Programming (MILP) solver to be used. If set to ``None``, the default one
      is used. For more information on MILP solvers and which default solver is
      used, see the method :meth:`solve
      <sage.numerical.mip.MixedIntegerLinearProgram.solve>` of the class
      :class:`MixedIntegerLinearProgram
      <sage.numerical.mip.MixedIntegerLinearProgram>`.

    - ``verbose`` -- integer (default: 0); sets the level of verbosity. Set
      to 0 by default, which means quiet.

    - ``integrality_tolerance`` -- float; parameter for use with MILP solvers
      over an inexact base ring; see
      :meth:`MixedIntegerLinearProgram.get_values`.

    EXAMPLES:

    Empty graph::

        sage: G = Graph()
        sage: G.maximum_leaf_number()
        0

    Petersen graph::

        sage: G = graphs.PetersenGraph()
        sage: G.maximum_leaf_number()                                                   # needs sage.numerical.mip
        6

    TESTS:

    One vertex::

        sage: G = Graph(1)
        sage: G.maximum_leaf_number()
        0

    Two vertices::

        sage: G = graphs.PathGraph(2)
        sage: G.maximum_leaf_number()
        1

    Unconnected graph::

        sage: G = Graph(2)
        sage: G.maximum_leaf_number()
        Traceback (most recent call last):
        ...
        ValueError: the graph must be connected
    """
    if G.order() <= 1:
        return 0
    if not G.is_connected():
        raise ValueError('the graph must be connected')
    if G.order() <= 3:
        return G.order() - 1
    return G.order() - dominating_set(G, connected=True, value_only=True,
                                      solver=solver, verbose=verbose,
                                      integrality_tolerance=integrality_tolerance)



# sage_setup: distribution = sagemath-graphs
# sage.doctest: needs sage.graphs
from sage.topology.simplicial_complex import SimplicialComplex, Simplex

from sage.topology.simplicial_complex_morphism import SimplicialComplexMorphism

from sage.topology.delta_complex import DeltaComplex, delta_complexes

from sage.topology.cubical_complex import CubicalComplex, cubical_complexes

from sage.misc.lazy_import import lazy_import
lazy_import('sage.topology.filtered_simplicial_complex', 'FilteredSimplicialComplex')

lazy_import('sage.topology', 'simplicial_complex_catalog', 'simplicial_complexes')
lazy_import('sage.topology', 'simplicial_set_catalog', 'simplicial_sets')

lazy_import('sage.topology.moment_angle_complex', 'MomentAngleComplex')

# # For taking care of old pickles
# from sage.misc.persist import register_unpickle_override
# register_unpickle_override('sage.topology.simplicial_complex_examples', 'SimplicialSurface', SimplicialComplex)
del lazy_import



# sage_setup: distribution = sagemath-graphs
from libc.stdint cimport uint8_t

cdef class FastDigraph:
    cdef uint8_t n
    cdef int * graph
    cdef list int_to_vertices
    cdef int * degree

cdef int compute_out_neighborhood_cardinality(FastDigraph, int) noexcept

cdef int popcount32(int) noexcept
cdef int slow_popcount32(int) noexcept



# sage_setup: distribution = sagemath-graphs
r"""
Interactive window for cluster seed mutations
"""

import ipywidgets as widgets
from sage.misc.latex import latex
from sage.repl.rich_output.pretty_print import pretty_print
from IPython.display import clear_output


def cluster_interact(self, fig_size=1, circular=True, kind='seed'):
    r"""
    Start an interactive window for cluster seed mutations.

    Only in *Jupyter notebook mode*.

    Not to be called directly. Use the :meth:`interact` methods
    of :class:`ClusterSeed` and :class:`ClusterQuiver` instead.

    INPUT:

    - ``fig_size`` -- (default: 1) factor by which the size of the
      plot is multiplied

    - ``circular`` -- boolean (default: ``True``); if ``True``, the circular
      plot is chosen, otherwise >>spring<< is used

    - ``kind`` -- either ``'seed'`` (default) or ``'quiver'``

    TESTS::

        sage: S = ClusterSeed(['A',4])                                                  # needs sage.graphs sage.modules
        sage: S.interact()   # indirect doctest                                         # needs sage.graphs sage.modules sage.symbolic
        ...VBox(children=...
    """
    if kind not in ['seed', 'quiver']:
        raise ValueError('kind must be "seed" or "quiver"')

    show_seq = widgets.Checkbox(value=True,
                                description="Display mutation sequence")

    show_vars = widgets.Checkbox(value=True,
                                 description="Display cluster variables")

    show_matrix = widgets.Checkbox(value=True,
                                   description="Display B-matrix")

    show_lastmutation = widgets.Checkbox(value=True,
                                         description="Show last mutation vertex")

    mut_buttons = widgets.ToggleButtons(options=list(range(self._n)),
                                       style={'button_width':'initial'},
                                       description='Mutate at: ')

    which_plot = widgets.Dropdown(options=['circular', 'spring'],
                                  value='circular' if circular else "spring",
                                  description='Display style:')

    out = widgets.Output()

    seq = []

    def print_data():
        if show_seq.value:
            pretty_print("Mutation sequence: ", seq)

        if show_vars.value and kind == 'seed':
            pretty_print("Cluster variables:")
            table = "\\begin{align*}\n"
            for i in range(self._n):
                table += "\tv_{%s} &= " % i + latex(self.cluster_variable(i)) + "\\\\ \\\\\n"
            table += "\\end{align*}"
            pretty_print(table)

        if show_matrix.value:
            pretty_print("B-Matrix: ", self._M)

    def refresh(w):
        k = mut_buttons.value
        circular = bool(which_plot.value == "circular")
        with out:
            clear_output(wait=True)
            if show_lastmutation.value:
                self.show(fig_size=fig_size, circular=circular, mark=k)
            else:
                self.show(fig_size=fig_size, circular=circular)
            print_data()

    def do_mutation(*args, **kwds):
        k = mut_buttons.value
        circular = bool(which_plot.value == "circular")
        self.mutate(k)
        if seq and seq[-1] == k:
            seq.pop()
        else:
            seq.append(k)
        with out:
            clear_output(wait=True)
            if show_lastmutation.value:
                self.show(fig_size=fig_size, circular=circular, mark=k)
            else:
                self.show(fig_size=fig_size, circular=circular)
            print_data()

    mut_buttons.on_msg(do_mutation)

    show_seq.observe(refresh, 'value')
    if kind == 'seed':
        show_vars.observe(refresh, 'value')
    show_matrix.observe(refresh, 'value')
    show_lastmutation.observe(refresh, 'value')
    which_plot.observe(refresh, 'value')

    mut_buttons.on_widget_constructed(refresh)

    if kind == 'seed':
        top = widgets.HBox([show_seq, show_vars])
    else:
        top = widgets.HBox([show_seq])

    return widgets.VBox([which_plot,
                         top,
                         widgets.HBox([show_matrix, show_lastmutation]),
                         mut_buttons, out])
